/*
 * This file and everything in this package is covered under
 * the GNU General Public License.  See the LICENSE file at the top
 * level of this package for full details.  A copy of the license
 * is also available at: http://opendj.org/docs/licenses/LICENSE.gpl
 *
 * Copyright 2001 DAX Interactive, LLC.
 * 
 * $Author: jonathan $
 * $Revision: 1.1.1.1 $
 * $Header: /home/cvspublic/opendj/opendj-plugins/src/org/opendj/plugins/powermon/PowerMonitor.java,v 1.1.1.1 2001/09/29 20:36:57 jonathan Exp $
 * $Log: PowerMonitor.java,v $
 * Revision 1.1.1.1  2001/09/29 20:36:57  jonathan
 * Imported opendj-plugins
 *
 */
package org.opendj.plugins.powermon;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import org.opendj.util.Util;
import org.opendj.util.DBUtil;
import org.opendj.util.ProcessUtil;
import org.opendj.util.Config;
import org.opendj.util.log.Logger;
import org.opendj.util.service.Service;
import org.opendj.util.service.ServiceException;
import org.opendj.util.debug;

public class PowerMonitor implements Service {

    private volatile Thread itsMainThread = null;
    private volatile boolean isDone = false;

    private String  itsOnBatteryCommand   = null;
    private String  itsBattChargeCommand  = null;
    private String  itsTimeLeftCommand    = null;
    private long    itsRegularSleepMillis = -1;
    private String  itsNormalCommand      = null;
    private String  itsLossCommand        = null;

    private long    itsSleepMillis        = -1;
    private boolean isFirstRun            = true;
    private boolean isInAlertMode         = false;
    private int     itsFileDeployed       = FILE_INIT;

    public static final int FILE_INIT      = 0;
    public static final int FILE_NORMAL    = 1;
    public static final int FILE_POWERLOSS = 2;

    public static void main ( String[] args ) {
        PowerMonitor p = new PowerMonitor();
        try {
            p.initService();
            p.run();

        } catch ( Exception e ) {
            System.err.println("Exception in PowerMonitor.go: " + e.toString());
            e.printStackTrace();

        } finally {
            try { p.stopService(); } catch ( Exception e ) {}
        }
    }

    public PowerMonitor () {}

    public void initService () throws ServiceException {
        try {
            Config aConfig = Config.global();
            itsOnBatteryCommand = aConfig.getProperty(Config.PROP_ONBATTERY_COMMAND);
            itsBattChargeCommand = aConfig.getProperty(Config.PROP_BATTCHARGE_COMMAND);
            itsTimeLeftCommand = aConfig.getProperty(Config.PROP_TIMELEFT_COMMAND);
            itsRegularSleepMillis
                = Util.MILLIS_IN_MINUTE
                * Long.parseLong(aConfig.getProperty(Config.PROP_POWERMONITOR_MINS));
            itsNormalCommand = aConfig.getProperty(Config.PROP_FILENORMAL_COMMAND);
            itsLossCommand = aConfig.getProperty(Config.PROP_FILELOSS_COMMAND);

        } catch ( Exception e ) {
            throw new ServiceException(e);
        }
    }

    public void run () {

        isDone = false;
        itsMainThread = Thread.currentThread();
        itsSleepMillis = itsRegularSleepMillis;
        isFirstRun = true;
        isInAlertMode = false;
        itsFileDeployed = FILE_INIT;

        while ( !isDone ) {
            getPowerStatus();
            Util.sleep(itsSleepMillis);
        }
    }

    public void stopService () {
        isDone = true;
        if ( itsMainThread != null ) {
            itsMainThread.interrupt();
        }
    }

    private void getPowerStatus () {
        try {
            String aTimeOnBatteryString = ProcessUtil.execStringCommand(itsOnBatteryCommand);
            String aBattChargeString    = ProcessUtil.execStringCommand(itsBattChargeCommand);
            String aTimeLeftString      = ProcessUtil.execStringCommand(itsTimeLeftCommand);

            int    aTimeOnBattery = Integer.parseInt(aTimeOnBatteryString.trim());
            double aBattCharge    = Double.parseDouble(aBattChargeString.trim());
            double aTimeLeft      = Double.parseDouble(aTimeLeftString.trim());
            
            debug.out("PowerMonitor.getPowerStatus: bcharge=" + aBattCharge
                      + ", timeleft=" + aTimeLeft);

            // Adjust for server shutdown delay
            aTimeLeft = Math.max(aTimeLeft - 5.0, 0.0);

            updateDBStats(aBattCharge, aTimeLeft);

            if ( aTimeOnBattery > 0 ) {
                isInAlertMode = true;
                itsSleepMillis = Util.MILLIS_IN_MINUTE;
                if ( itsFileDeployed != FILE_POWERLOSS ) {
                    // Copy alert file
                    ProcessUtil.execWithWait(itsLossCommand);
                    itsFileDeployed = FILE_POWERLOSS;
                }

            } else {
                isInAlertMode = false;
                itsSleepMillis = itsRegularSleepMillis;
                if ( itsFileDeployed != FILE_NORMAL ) {
                    // Copy normal file
                    ProcessUtil.execWithWait(itsNormalCommand);
                    itsFileDeployed = FILE_NORMAL;
                }
            }
        } catch ( Exception e ) {
            Logger.global().capture("PowerMonitor.getPowerStatus", e, Logger.L_ERROR);
        }
    }

    private void updateDBStats ( double theBattCharge, double theTimeLeft ) throws SQLException {
        Connection aConn = null;
        PreparedStatement aPS = null;
        try {
            aConn = DBUtil.getConnection();
            if ( isFirstRun ) {
                aConn.setAutoCommit(false);
                aPS = aConn.prepareStatement("DELETE FROM PowerStatus");
                aPS.executeUpdate();
                aPS = aConn.prepareStatement
                    ("INSERT INTO PowerStatus ( batterycharge, timeleft ) VALUES ( ?, ? )");
                aPS.setDouble(1, theBattCharge);
                aPS.setDouble(2, theTimeLeft);
                aPS.executeUpdate();
                aConn.commit();
                isFirstRun = false;
                debug.out("Inserted " + theBattCharge + ", " + theTimeLeft + " into DB");

            } else {
                aPS = aConn.prepareStatement
                    ("UPDATE PowerStatus SET batterycharge = ?, timeleft = ?");
                aPS.setDouble(1, theBattCharge);
                aPS.setDouble(2, theTimeLeft);
                aPS.executeUpdate();
            }

        } finally {
            DBUtil.closeStatement(aPS);
            DBUtil.closeConnection(aConn);
        }
    }

    public static PowerStatus getPowerStatusForClient () {
        Connection aConn = null;
        PreparedStatement aPS = null;
        ResultSet aRS = null;
        try {
            aConn = DBUtil.getConnection();
            aPS = aConn.prepareStatement("SELECT batterycharge, timeleft FROM PowerStatus");
            aRS = aPS.executeQuery();
            if ( aRS.next() ) {
                return new PowerStatus( aRS.getDouble(1), aRS.getDouble(2) );
            } else {
                Logger.global().capture("PowerMonitor.getBatteryCharge: no status in DB", 
                                        null,
                                        Logger.L_ERROR);
                return null;
            }

        } catch ( Exception e ) {
            Logger.global().capture("PowerMonitor.getBatteryCharge", e, Logger.L_ERROR);
            return null;

        } finally {
            DBUtil.closeStatement(aPS);
            DBUtil.closeConnection(aConn);
        }
    }

}
