/*
 * This file and everything in this package is covered under
 * the GNU General Public License.  See the LICENSE file at the top
 * level of this package for full details.  A copy of the license
 * is also available at: http://opendj.org/docs/licenses/LICENSE.gpl
 *
 * Copyright 2001 DAX Interactive, LLC.
 * 
 * $Author: jonathan $
 * $Revision: 1.1.1.1 $
 * $Header: /home/cvspublic/opendj/opendj-plugins/src/org/opendj/plugins/rerun/player/RerunHandler.java,v 1.1.1.1 2001/09/29 20:36:57 jonathan Exp $
 * $Log: RerunHandler.java,v $
 * Revision 1.1.1.1  2001/09/29 20:36:57  jonathan
 * Imported opendj-plugins
 *
 */
package org.opendj.plugins.rerun.player;

import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.BufferedReader;
import java.io.IOException;
import java.net.Socket;
import java.sql.Connection;
import java.util.StringTokenizer;
import java.util.List;
import java.util.Vector;
import java.util.Iterator;
import org.opendj.util.Util;
import org.opendj.util.DBUtil;
import org.opendj.util.Config;
import org.opendj.util.debug;
import org.opendj.util.log.Logger;
import org.opendj.server.DJConnection; // just for protocol constants
import org.opendj.beans.archive.AudioArchiveBean;
import org.opendj.beans.onair.OnAirBean;

public class RerunHandler implements Runnable {

    public static final int STATE_WAITING      = 0;
    public static final int STATE_BEFORE_START = 1;
    public static final int STATE_BROADCASTING = 2;
    public static final int STATE_BEFORE_END   = 3;
    public static final int STATE_DONE         = 4;
    public static final int STATE_ERROR        = 5;

    // This MUST be less than LIMIT_CONNECT_TO_SERVER_MILLIS
    public static final long SLEEP_WAITING      = Util.MILLIS_IN_MINUTE;

    // This MUST be less than 30 seconds
    // It is inextricably tied to the PStore.RERUN_ADV_MILLIS constant,
    // which must be a safely larger value than this.
    public static final long SLEEP_BEFORE_START = 20000;

    public static final long SLEEP_BROADCASTING = 15000;
    public static final long SLEEP_BEFORE_END = 5000;

    private volatile Thread itsMainThread = null;
    private volatile boolean isDone = false;
    private volatile boolean endedOK = false;

    // Config constants
    private long   LIMIT_CONNECT_TO_SERVER_MILLIS = -1;
    private long   LIMIT_STREAM_SETUP_MILLIS = -1;
    private long   ASK_INTERVAL = -1;
    private String DJSERVER_HOST = null;
    private int    DJSERVER_PORT = -1;
    private String RERUN_EMAIL = null;
    private String RERUN_PASSWORD = null;

    private List itsOnAirBeans = null;
    private OnAirBean itsActiveBean = null;
    private int itsState = STATE_WAITING;
    private Logger LOG = null;
    private long itsLastAskTime = 0;

    private Socket itsSocket = null;
    private BufferedReader itsInReader  = null;
    private PrintWriter    itsOutWriter = null;

    private RerunStreamer itsStreamer = null;
    private Thread itsStreamerThread = null;

    public RerunHandler ( OnAirBean theOnAirBean, Logger theLogger ) {

        itsOnAirBeans = new Vector();
        synchronized (itsOnAirBeans) {
            itsOnAirBeans.add(theOnAirBean);
        }
        itsActiveBean = theOnAirBean;
        LOG = theLogger;

        Config aConfig = Config.global();
        LIMIT_CONNECT_TO_SERVER_MILLIS 
            = Long.parseLong(aConfig.getProperty(Config.PROP_LIMITCONNECTTOSERVER));
        LIMIT_STREAM_SETUP_MILLIS
            = Long.parseLong(aConfig.getProperty(Config.PROP_LIMITSTREAMSETUP));
        ASK_INTERVAL
            = Long.parseLong(aConfig.getProperty(Config.PROP_RERUNASKINTERVAL));
        DJSERVER_HOST  = aConfig.getProperty(Config.PROP_DJSERVER_IP);
        DJSERVER_PORT  = Integer.parseInt(aConfig.getProperty(Config.PROP_DJSERVER_PORT));
        RERUN_EMAIL    = aConfig.getProperty(Config.PROP_RERUNDJEMAIL);
        RERUN_PASSWORD = aConfig.getProperty(Config.PROP_RERUNDJPASSWORD);
    }

    public void run () {

        itsMainThread = Thread.currentThread();
        isDone = false;
        endedOK = false;
        itsState = STATE_WAITING;
        long aSleepInterval = -1;

        try {
            while ( !isDone ) {
                switch ( itsState ) {
                case STATE_WAITING:
                    aSleepInterval = doWaiting();
                    break;

                case STATE_BEFORE_START:
                    aSleepInterval = doBeforeStart();
                    break;

                case STATE_BROADCASTING:
                    aSleepInterval = doBroadcasting();
                    break;

                case STATE_BEFORE_END:
                    aSleepInterval = doBeforeEnd();
                    break;

                case STATE_DONE:
                    debug.out("RerunHandler: in STATE_DONE");
                    cleanup();
                    synchronized (itsOnAirBeans) {
                        itsOnAirBeans.remove(0);
                        if ( itsOnAirBeans.size() > 0 ) {
                            debug.out("RerunHandler: in STATE_DONE - more programs, "
                                      + "re-entering WAITING state");
                            itsActiveBean = (OnAirBean) itsOnAirBeans.get(0);
                            itsState = STATE_WAITING;
                            aSleepInterval = 0;
                            break;
                        } else {
                            debug.out("RerunHandler: in STATE_DONE - no more programs");
                            endedOK = true;
                            return;
                        }
                    }

                default:
                    // Whoa -- could be STATE_ERROR
                    log("RerunHandler.run: unexpected state (probably STATE_ERROR)");
                    LOG.capture("Unexpected state: " + String.valueOf(itsState), 
                                null, Logger.L_FIRE);
                    return;
                }
                Util.sleep(aSleepInterval);
            }
        } finally {
            cleanup();
        }
    }

    public boolean isDone () { return isDone; }

    public boolean endedOK () { return endedOK; }

    public void finish () {
        isDone = true;
        if ( itsMainThread != null ) {
            itsMainThread.interrupt();
            itsMainThread = null;
        }
    }

    public OnAirBean    getOnAirBean () { return itsActiveBean; }
    public Logger       getLogger    () { return LOG; }

    public boolean hasPredecessorTo ( OnAirBean theOnAirBean ) {

        OnAirBean anOnAirBean = null;
        synchronized (itsOnAirBeans) {
            Iterator i = itsOnAirBeans.iterator();
            while ( i.hasNext() ) {
                anOnAirBean = (OnAirBean) i.next();
                if ( anOnAirBean.isPredecessorTo(theOnAirBean) ) {
                    return true;
                }
            }
        }
        return false;
    }

    public void addProgram ( OnAirBean theOnAirBean ) {
        synchronized (itsOnAirBeans) { 
            itsOnAirBeans.add(theOnAirBean);
        }
    }

    private void cleanup () {
        debug.out("RerunHandler.cleanup called from: " 
                  + Util.getStackTrace(new Exception("trace")));
        if ( itsSocket != null ) {
            try {
                sendData(DJConnection.REQ_DISCONNECT);
                String aServerResponse = readData();
                if ( !aServerResponse.startsWith(DJConnection.RES_BYE) ) {
                    LOG.capture("Error logging out: '" + aServerResponse + "'", null, Logger.L_WARNING);
                }
            } catch ( Exception e ) {}
        }
        try { itsSocket.close(); } catch ( Exception e ) {}
        itsSocket = null;
        itsInReader = null;
        itsOutWriter = null;
        stopStreamer();
    }

    private void stopStreamer () {
        if ( itsStreamer != null ) {
            itsStreamer.finish();
            // Give the streamer a bit to wrap up, then kill it
            if ( itsStreamerThread.isAlive() ) {
                try { itsStreamerThread.join(5000); } catch ( Exception e ) {}
                if ( itsStreamerThread.isAlive() ) {
                    LOG.capture("RerunHandler.cleanup: Streamer did not exit gracefully, "
                                + "terminating with Thread.stop", null, Logger.L_ERROR);
                    itsStreamerThread.stop();
                }
            }
            itsStreamer = null;
        }
    }

    private long doWaiting () { 

        // Calculate how long before we're supposed to start
        long aDelta = itsActiveBean.getstarttime() - System.currentTimeMillis();

        debug.out("RerunHandler.doWaiting: delta=" + aDelta);

        // Oops! Start now!
        if ( aDelta < 0 ) {
            connectToServer();
            return startBroadcasting();
        }

        // Are we close enough that we should connect to the server?
        if ( aDelta < LIMIT_CONNECT_TO_SERVER_MILLIS ) {
            connectToServer();
            return 0;
        }

        return SLEEP_WAITING;
    }

    private long doBeforeStart () { 
        return startBroadcasting();
    }

    private long doBroadcasting () { 
        // Make sure our streamer is still OK
        if ( !itsStreamer.isStreaming() ) {
            debug.out("RerunHandler.doBroadcasting: streamer died!");
            LOG.capture("RerunHandler.doBroadcasting: Ack! Streamer died!", null, Logger.L_FIRE);
            itsState = STATE_ERROR;
            cleanup();
            return 0;
        }

        // Streamer is still OK, should we ask the server 
        // how much longer we have to go?
        long aNowTime = System.currentTimeMillis();
        if ( aNowTime - itsLastAskTime > ASK_INTERVAL ) {
            if ( checkIsStreamDone() ) {
                return 0;
            }
            itsLastAskTime = aNowTime;
        }

        return SLEEP_BROADCASTING;
    }

    private long doBeforeEnd () {
        if ( checkIsStreamDone() ) {
            return 0;
        }
        return SLEEP_BEFORE_END;
    }

    // Connect the socket, do basic authentication
    private void connectToServer () {

        debug.out("RerunHandler.connectToServer...");
        String anAuthCommand = DJConnection.REQ_AUTH + " " + RERUN_EMAIL + " " + RERUN_PASSWORD;
        String aServerResponse = null;

        try {
            itsSocket = new Socket(DJSERVER_HOST, DJSERVER_PORT);
            itsInReader = new BufferedReader
                ( new InputStreamReader(itsSocket.getInputStream()));
            itsOutWriter = new PrintWriter
                ( new OutputStreamWriter(itsSocket.getOutputStream()), true);

            sendData(anAuthCommand);
            aServerResponse = readData();
            if ( aServerResponse.startsWith(DJConnection.RES_AUTHOK) ) {
                debug.out("RerunHandler.connectToServer connect OK");
                itsState = STATE_BEFORE_START;
            } else {
                throw new RerunPlayerException(RerunPlayerException.E_AUTH_FAILED
                                               + ": '" + aServerResponse + "'");
            }

        } catch ( Exception e ) {
            LOG.capture("RerunHandler.connectToServer", e, Logger.L_FIRE);
            itsState = STATE_ERROR;
            cleanup();
        }
    }

    // Ask the server if we can start broadcasting now
    private long startBroadcasting () {

        debug.out("RerunHandler.startBroadcasting...");
        String aServerResponse = null;
        String aBroadcastCommand = DJConnection.REQ_BROADCAST + " " + itsActiveBean.getchannelpath();

        try {
            sendData(aBroadcastCommand);
            aServerResponse = readData();

            if ( aServerResponse.startsWith(DJConnection.RES_GO)  ) {
                // Now let's get rolling!
                debug.out("RerunHandler.startBroadcasting: server said GO");
                log("server says GO");
                return reallyStartBroadcasting();
            }
            if ( aServerResponse.startsWith(DJConnection.RES_DELAY) ) {
                StringTokenizer st = new StringTokenizer(aServerResponse);
                st.nextToken(); // skip the command
                long aDelay = Long.parseLong(st.nextToken());
            
                debug.out("RerunHandler.startBroadcasting: server said " + aServerResponse);

                // If the delay is less than 30 seconds, accept it and delay
                // but decrease the delay by 3 seconds so we don't miss our window
                // Otherwise return false
                if ( aDelay < (30*1000) ) {
                    log("delay is less than 30 seconds, sleeping for (" 
                        + String.valueOf(aDelay - (3*1000)) + "), then asking again...");
                    Util.sleep(Math.max(aDelay - (3*1000), 0));
                    return startBroadcasting();

                } else {
                    // Delay is too long, give up
                    log("delay is too long, returning SLEEP_BEFORE_START");
                    return SLEEP_BEFORE_START;
                }
            } else {
                throw new RerunPlayerException(RerunPlayerException.E_BROADCAST_FAILED 
                                               + ": '" + aServerResponse + "'");
            }

        } catch ( Exception e ) {
            LOG.capture("RerunHandler.startBroadcasting", e, Logger.L_FIRE);
            itsState = STATE_ERROR;
            cleanup();
            return 0;
        }
    }

    private long reallyStartBroadcasting () {

        debug.out("RerunHandler.reallyStartBroadcasting...");

        // Sleep for a bit to give icecast a chance to
        // notice that the last program has disconnected
        Util.sleep(250);

        // Start a new streamer
        itsStreamer = new RerunStreamer(this);
        itsStreamerThread = new Thread(itsStreamer);
        itsStreamerThread.start();

        // Make sure the streamer connected OK
        long aMarkTime = System.currentTimeMillis();
        boolean aSetupFailed = false;
        while ( !itsStreamer.isStreaming() ) {
            if (System.currentTimeMillis() - aMarkTime > LIMIT_STREAM_SETUP_MILLIS) {
                aSetupFailed = true;
                break;
            }
            Util.sleep(250);
        }
        if ( aSetupFailed ) {
            LOG.capture("RerunHandler.reallyStartBroadcasting: streamer failed to setup", 
                        null, Logger.L_FIRE);
            itsState = STATE_ERROR;
            cleanup();
            return 0;
        }

        // Set the state
        itsState = STATE_BROADCASTING;

        // While broadcasting, we'll watch the streamer every 15 seconds
        return SLEEP_BROADCASTING;
    }

    private boolean checkIsStreamDone () {
        String aServerResponse = null;
        debug.out("RerunHandler.checkIsStreamDone called");
        try {
            sendData(DJConnection.REQ_WAITFOREXIT);
            aServerResponse = readData();
            if ( aServerResponse.startsWith(DJConnection.RES_CLOSENOW) ) {
                debug.out("RerunHandler.checkIsStreamDone: server said CLOSENOW");
                itsState = STATE_DONE;
                return true;
            }
            if ( aServerResponse.startsWith(DJConnection.RES_DELAY) ) {
                debug.out("RerunHandler.checkIsStreamDone: server said " + aServerResponse);
                StringTokenizer st = new StringTokenizer(aServerResponse);
                st.nextToken(); // skip the command
                long aDelay = Long.parseLong(st.nextToken());
                if ( aDelay < (2*SLEEP_BEFORE_END) ) {
                    Util.sleep(aDelay);
                    itsState = STATE_DONE;
                    return true;
                }
                if ( itsState != STATE_BEFORE_END && aDelay < ASK_INTERVAL ) {
                    itsState = STATE_BEFORE_END;
                    return true;
                }
                return false;
            }
            if ( aServerResponse.startsWith(DJConnection.RES_EXPIRED) ) {
                debug.out("RerunHandler.checkIsStreamDone: server said EXPIRED");
                // This means that the server has already disconnected us
                // No reason to continue broadcasting
                itsState = STATE_DONE;
                return true;
            } else {
                throw new RerunPlayerException(RerunPlayerException.E_WAITFOREXIT_FAILED 
                                               + ": '" + aServerResponse + "'");
            }
        } catch ( Exception e ) {
            LOG.capture("Unexpected exception", e, Logger.L_FIRE);
            itsState = STATE_ERROR;
            return true;
        }
    }

    public AudioArchiveBean loadArchive (OnAirBean theOnAirBean) {
        // Load archive corresponding to the onair bean
        AudioArchiveBean anArchive = new AudioArchiveBean();
        Connection aConn = null;
        try {
            aConn = DBUtil.getConnection();
            anArchive.setArchid(theOnAirBean.getarchid());
            anArchive.lookupByArchID(aConn);
            return anArchive;

        } catch ( Exception e ) {
            LOG.capture("Error loading AudioArchiveBean", e, Logger.L_FIRE);
            return null;

        } finally {
            DBUtil.closeConnection(aConn);
        }
    }

    private void sendData ( String s ) {
        log(">" + s);
        itsOutWriter.println(s);
    }

    private String readData () throws IOException {
        String aReadData = itsInReader.readLine(); 
        log("<" + aReadData);
        return aReadData;
    }

    private void log ( String s ) {
        debug.out(s);
        // LOG.capture(s, null, Logger.L_DEBUG);
    }
}
