/*
 * This file and everything in this package is covered under
 * the GNU General Public License.  See the LICENSE file at the top
 * level of this package for full details.  A copy of the license
 * is also available at: http://opendj.org/docs/licenses/LICENSE.gpl
 *
 * Copyright 2001 DAX Interactive, LLC.
 * 
 * $Author: jonathan $
 * $Revision: 1.1.1.1 $
 * $Header: /home/cvspublic/opendj/opendj-plugins/src/org/opendj/plugins/rerun/player/RerunPlayer.java,v 1.1.1.1 2001/09/29 20:36:57 jonathan Exp $
 * $Log: RerunPlayer.java,v $
 * Revision 1.1.1.1  2001/09/29 20:36:57  jonathan
 * Imported opendj-plugins
 *
 */
package org.opendj.plugins.rerun.player;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.Vector;
import org.opendj.util.Util;
import org.opendj.util.DBUtil;
import org.opendj.util.Config;
import org.opendj.util.debug;
import org.opendj.util.log.Logger;
import org.opendj.util.service.Service;
import org.opendj.util.service.ServiceException;
import org.opendj.beans.onair.OnAirBean;

public class RerunPlayer implements Service {

    public static final long SLEEP_INTERVAL = ( Util.MILLIS_IN_MINUTE );
    public static final String LOGFILE_NAME = "RerunPlayer";

    private volatile Thread itsMainThread = null;
    private volatile boolean isDone = false;
    private Logger LOG = null;
    private Map itsHandledPrograms = null;
    private int RERUN_DJID = -1;

    public static void main ( String[] args ) {
        RerunPlayer r = new RerunPlayer();
        try {
            r.initService();
            r.run();
        } catch ( Exception e ) {
            System.err.println("Exception in RerunPlayer.run: " + e.toString());
            e.printStackTrace();
        } finally {
            r.stopService();
        }
    }

    public RerunPlayer () {}

    public void run () {

        itsMainThread = Thread.currentThread();
        isDone = false;
        OnAirBean anOnAirBean = null;

        // Give the AudioServer and AudioChannels time to setup
        Util.sleep(15000);

        try {
            while ( !isDone ) {
                // Is there any program on soon enough that we should
                // worry about it?
                try {
                    checkPrograms();
                    
                    anOnAirBean = getNextProgram();
                    if ( anOnAirBean != null ) {
                        handleProgram(anOnAirBean);
                        
                    } else {
                        // Only sleep if we didn't find anything.  This way
                        // if there are multiple programs about to start on
                        // different channels, we'll see them all before sleeping.
                        debug.out("sleeping for " + SLEEP_INTERVAL);
                        Util.sleep(SLEEP_INTERVAL);
                    }
                    
                } catch ( Exception e ) {
                    LOG.capture("Unexpected exception in RerunPlayer.run", e, Logger.L_FIRE);
                }
            }
        } finally {
            try {
                stopPrograms();
            } catch ( Exception e ) {
                LOG.capture("RerunPlayer.run, exception in stopPrograms", e, Logger.L_FIRE);
            }
        }
    }

    public void initService () throws ServiceException {
        try {
            LOG = new Logger(LOGFILE_NAME);

            Config aConfig = Config.global();
            RERUN_DJID = Integer.parseInt(aConfig.getProperty(Config.PROP_RERUNDJID));

            itsHandledPrograms = new HashMap();

        } catch ( Exception e ) {
            throw new ServiceException(e);
        }
    }

    public void stopService () {
        isDone = true;
        if ( itsMainThread != null ) {
            itsMainThread.interrupt();
            itsMainThread = null;
        }
    }

    private OnAirBean getNextProgram () throws SQLException {
        Connection aConn = null;
        PreparedStatement aPS = null;
        ResultSet aRS = null;
        String aSQL
            = "SELECT * "
            + "FROM PublishedSchedule "
            + "WHERE endtime > ? "
            + "AND starttime < ? "
            + "AND djid = ? "
            + "ORDER BY starttime "
            ;
        long aNowTime = System.currentTimeMillis();
        long aStartLookahead = aNowTime + (Util.MILLIS_IN_MINUTE * 30);
        OnAirBean anOnAirBean = null;

        try {
            aConn = DBUtil.getConnection();
            aPS = aConn.prepareStatement(aSQL);
            int i=1;
            aPS.setLong(i++, aNowTime);
            aPS.setLong(i++, aStartLookahead);
            aPS.setInt (i++, RERUN_DJID);
            aRS = aPS.executeQuery();
            while ( aRS.next() ) {
                anOnAirBean = new OnAirBean();
                anOnAirBean.init(aRS);
                if ( hasBeenHandled(anOnAirBean.getpsid()) ) {
                    continue;
                }
                return anOnAirBean;
            }

        } finally {
            DBUtil.closeRS(aRS);
            DBUtil.closeStatement(aPS);
            DBUtil.closeConnection(aConn);
        }

        debug.out("RerunPlayer.run: no next program found");
        return null;
    }

    private boolean hasBeenHandled ( int thePSID ) {
        Integer aPSIDInteger = new Integer(thePSID);
        return ( itsHandledPrograms.get(aPSIDInteger) != null );
    }

    private void handleProgram ( OnAirBean theOnAirBean ) {

        ProgramData aProgData = null;

        try {
            debug.out("handleProgram(" + theOnAirBean + ") called");

            // Is this contiguous with any broadcasts in progress?
            Iterator i = itsHandledPrograms.keySet().iterator();
            while ( i.hasNext() ) {
                aProgData = (ProgramData) itsHandledPrograms.get(i.next());
                if ( aProgData.handler.hasPredecessorTo(theOnAirBean) ) {
                    debug.out("handleProgram: appending " + theOnAirBean + " to existing handler");
                    aProgData.handler.addProgram(theOnAirBean);
                    itsHandledPrograms.put(new Integer(theOnAirBean.getpsid()), aProgData);
                    return;
                }
            }

            // Create a handler for this onair bean
            RerunHandler aHandler = new RerunHandler(theOnAirBean, LOG);
            Thread aHandlerThread = new Thread(aHandler);

            aProgData = new ProgramData( theOnAirBean, aHandler, aHandlerThread );

            itsHandledPrograms.put(new Integer(theOnAirBean.getpsid()), aProgData);
            aHandlerThread.start();

        } catch ( Exception e ) {
            LOG.capture("RerunPlayer.handleProgram: unexpected exception", e, Logger.L_ERROR);
        }
    }

    // Check all programs in progress, make sure they're still running OK
    private void checkPrograms () {

        ProgramData aProgData = null;
        List someProgsToRemove = new Vector();
        Integer aPSIDInteger = null;

        Iterator i = itsHandledPrograms.keySet().iterator();
        while ( i.hasNext() ) {
            aPSIDInteger = (Integer) i.next();
            aProgData = (ProgramData) itsHandledPrograms.get(aPSIDInteger);

            if ( aProgData.thread == null || 
                 !aProgData.thread.isAlive() ||
                 aProgData.handler == null ||
                 aProgData.handler.isDone() ) {

                if ( aProgData.handler.endedOK() ) {
                    debug.out("Handler for " + aProgData.handler.getOnAirBean() + " ended OK, removing");
                    // Handler ended OK, remove from map
                    someProgsToRemove.add(aPSIDInteger);
                    continue;
                }

                LOG.capture("RerunPlayer.checkPrograms: program " + aProgData + " has died, restarting.", 
                            null, Logger.L_ERROR);
                aProgData.handler = new RerunHandler(aProgData.oabean, LOG);
                aProgData.thread = new Thread(aProgData.handler);
                aProgData.thread.start();
            }
        }

        i = someProgsToRemove.iterator();
        while ( i.hasNext() ) {
            itsHandledPrograms.remove(i.next());
        }
    }

    // Stop all running programs
    private void stopPrograms () {
        
        Integer aPSIDInteger = null;
        ProgramData aProgData = null;

        Iterator i = itsHandledPrograms.keySet().iterator();
        while ( i.hasNext() ) {
            try {
                aPSIDInteger = (Integer) i.next();
                aProgData = (ProgramData) itsHandledPrograms.get(aPSIDInteger);
                aProgData.handler.finish();
                if ( aProgData.thread.isAlive() ) {
                    try {
                        aProgData.thread.join(2000);
                    } catch ( Exception e ) {}
                    if ( aProgData.thread.isAlive() ) {
                        LOG.capture("RerunPlayer.stopPrograms: Program " + aProgData
                                    + " did not end properly, terminating with Thread.stop",
                                    null, Logger.L_ERROR);
                        aProgData.thread.stop();
                    }
                }
            } catch ( Exception e ) {
                LOG.capture("RerunPlayer.stopPrograms: unexpected exception", e, Logger.L_FIRE); 
            }
        }
        itsHandledPrograms.clear();
    }

    public class ProgramData {

        public OnAirBean oabean = null;
        public RerunHandler handler = null;
        public Thread thread = null;

        public ProgramData ( OnAirBean theOnAirBean, 
                             RerunHandler theHandler, 
                             Thread theThread ) {
            oabean = theOnAirBean;
            handler = theHandler;
            thread = theThread;
        }

        public String toString () {
            return "[ProgData psid=" + oabean.getpsid() + ", channel='" 
                + oabean.getchannelpath() + "']";
        }
    }
}
