/*
 * This file and everything in this package is covered under
 * the GNU General Public License.  See the LICENSE file at the top
 * level of this package for full details.  A copy of the license
 * is also available at: http://opendj.org/docs/licenses/LICENSE.gpl
 *
 * Copyright 2001 DAX Interactive, LLC.
 * 
 * $Author: jonathan $
 * $Revision: 1.1.1.1 $
 * $Header: /home/cvspublic/opendj/opendj-plugins/src/org/opendj/plugins/rerun/scheduler/RerunScheduler.java,v 1.1.1.1 2001/09/29 20:36:57 jonathan Exp $
 * $Log: RerunScheduler.java,v $
 * Revision 1.1.1.1  2001/09/29 20:36:57  jonathan
 * Imported opendj-plugins
 *
 */
package org.opendj.plugins.rerun.scheduler;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import java.util.Map;
import java.util.HashMap;
import org.opendj.beans.archive.AudioArchiveBean;
import org.opendj.util.Util;
import org.opendj.util.DBUtil;
import org.opendj.util.Config;
import org.opendj.util.SchedBlock;
import org.opendj.util.debug;
import org.opendj.util.log.Logger;
import org.opendj.util.service.Service;
import org.opendj.util.service.ServiceException;

public class RerunScheduler implements Service {

    public static final long SLEEP_INTERVAL = ( Util.MILLIS_IN_HOUR * 6 );
    public static final String LOGFILE_NAME = "RerunScheduler";
    
    private volatile Thread itsMainThread = null;
    private volatile boolean isDone = false;
    private Logger LOG = null;

    private int  MAX_AVAIL_RERUNS         = 0;
    private long LASTAIR_LIMIT_HOURS      = 0;
    private int  MIN_KARMA_FOR_RERUN      = 0;
    private int  RERUN_DJID               = 0;
    private long RERUN_SCHED_LIMIT_MILLIS = 0;

    private int[] itsChannelids = null;
    private List itsRerunList = null;
    private Map itsBaseRerunMap = null;
    private Map itsActiveRerunMap = null;
    private RerunSchedulerStrategy itsStrategy = null;

    public static void main ( String[] args ) {
        RerunScheduler r = new RerunScheduler();
        try {
            r.initService();
            r.run();
        } catch ( Exception e ) {
            System.err.println("Exception in RerunScheduler.run: " + e.toString());
            e.printStackTrace();
        } finally {
            r.stopService();
        }
    }

    public RerunScheduler () {}

    public void run () {

        itsMainThread = Thread.currentThread();
        isDone = false;

        while ( !isDone ) {
            Connection aConn = null;
            int aChannelid = -1;

            // Assign appropriate DJ-specified reruns to the "rerun" DJ
            assignReruns();

            try {
                // For each channel...
                for ( int i=0; i<itsChannelids.length; i++ ) {

                    aChannelid = itsChannelids[i];
                    aConn = DBUtil.getConnection();

                    // Get a list of eligible reruns
                    getRerunsAvailable(aConn, aChannelid);

                    // Only continue if we actually have any reruns
                    if ( itsRerunList == null || itsRerunList.size() <= 0 ) {
                        debug.out("No candidate reruns found for channelid " + aChannelid);
                        continue;
                    }
                        
                    // Get a list of empty slots in the near future
                    List someEmptySlots = getEmptySlots(aConn, aChannelid);

                    // Only continue if there are actually any empty slots
                    if ( someEmptySlots == null || someEmptySlots.size() <= 0 ) {
                        debug.out("No empty slots found for channelid " + aChannelid);
                        continue;
                    }
            
                    // Iterate over all the empty slots until we've filled them all
                    int stop = someEmptySlots.size();
                    for ( int j=0; j<stop; j++ ) {
                        // Get the empty slot
                        EmptySlot aSlot = (EmptySlot) someEmptySlots.get(j);
                            
                        fillSlotWithReruns( aConn, aSlot ); 
                    }
                }

            } catch ( Exception e ) {
                debug.out("Unexpected exception in RerunScheduler.run: " + Util.getStackTrace(e));
                LOG.capture("Unexpected exception in RerunScheduler.run", e, Logger.L_ERROR);

            } finally {
                DBUtil.closeConnection(aConn);
            }

            // Wait several hours until we do this again
            Util.sleep(SLEEP_INTERVAL);
        }
    }

    public void initService () throws ServiceException {
        try {
            LOG = new Logger(LOGFILE_NAME);

            Config aConfig = Config.global();

            MAX_AVAIL_RERUNS
                = Integer.parseInt(aConfig.getProperty(Config.PROP_MAXAVAILRERUNS));
            LASTAIR_LIMIT_HOURS
                = Long.parseLong  (aConfig.getProperty(Config.PROP_LASTAIRLIMITHOURS));
            MIN_KARMA_FOR_RERUN
                = Integer.parseInt(aConfig.getProperty(Config.PROP_MINRERUNKARMA));
            RERUN_DJID
                = Integer.parseInt(aConfig.getProperty(Config.PROP_RERUNDJID));
            RERUN_SCHED_LIMIT_MILLIS
                = Long.parseLong  (aConfig.getProperty(Config.PROP_RERUNSCHEDLIMITHOURS))
                * Util.MILLIS_IN_HOUR;

            initChannelids();

            itsStrategy = new RerunSchedulerStrategy();

        } catch ( Exception e ) {
            throw new ServiceException(e);
        }
    }

    public void stopService () {
        isDone = true;
        if ( itsMainThread != null ) {
            itsMainThread.interrupt();
            itsMainThread = null;
        }
    }

    private void initChannelids () throws SQLException {

        List someChannelids = new Vector();
        Connection aConn = null;
        PreparedStatement aPS = null;
        ResultSet aRS = null;
        String aSQL = "SELECT cid FROM Channel ORDER BY cid";

        try {
            aConn = DBUtil.getConnection();
            aPS = aConn.prepareStatement(aSQL);
            aRS = aPS.executeQuery();
            while ( aRS.next() ) {
                someChannelids.add(new Integer(aRS.getInt(1)));
            }
        } finally {
            DBUtil.closeRS(aRS);
            DBUtil.closeStatement(aPS);
            DBUtil.closeConnection(aConn);
        }

        itsChannelids = new int[someChannelids.size()];
        for ( int i=0; i<itsChannelids.length; i++ ) {
            itsChannelids[i] = ((Integer) someChannelids.get(i)).intValue();
            debug.out("Added channelid: " + itsChannelids[i]);
        }
    }

    private void getRerunsAvailable ( Connection theConn,
                                      int theChannelid ) throws SQLException {
        itsRerunList = new Vector();

        PreparedStatement aPS = null;
        ResultSet aRS = null;
        String aSQL 
            = "SELECT * "
            + "FROM ArchivedShow "
            + "WHERE karmaearned > ? "
            + "AND iscompleted = 1 "
            + "AND isrerunnable = 1 "
            + "AND lastaired < ? "
            + "AND channelid = ? "
            + "ORDER BY karmaearned DESC, lastaired "
            + "LIMIT " + MAX_AVAIL_RERUNS + " "
            ;

        long aLastAirTimeLimit
            = System.currentTimeMillis() 
            - (Util.MILLIS_IN_HOUR * LASTAIR_LIMIT_HOURS);
        AudioArchiveBean aRerun = null;

        try {
            aPS = theConn.prepareStatement(aSQL);
            int i=1;
            aPS.setInt (i++, MIN_KARMA_FOR_RERUN);
            aPS.setLong(i++, aLastAirTimeLimit);
            aPS.setInt (i++, theChannelid);

            aRS = aPS.executeQuery();

            while ( aRS.next() ) {
                aRerun = new AudioArchiveBean();
                aRerun.init(aRS);

                // Does the rerun have a standard time?
                if ( aRerun.getScheduledDuration() % ( Util.MILLIS_IN_MINUTE * 30 ) != 0 ) {
                    debug.out("Skipping rerun with nonstandard duration: " + aRerun);
                    continue;
                }

                itsRerunList.add(aRerun);
                debug.out("Found candidate rerun: " + aRerun);
            }

        } finally {
            DBUtil.closeRS(aRS);
            DBUtil.closeStatement(aPS);
        }

        // Create rerun map
        if ( itsRerunList != null && itsRerunList.size() > 0 ) {
            createRerunMap();
        }
    }

    private void createRerunMap () {

        List aCommonTimeList = null;
        Iterator iter = itsRerunList.iterator();
        AudioArchiveBean aRerun = null;
        Long aNumHalfHours = null;
        itsActiveRerunMap = new HashMap();
        itsBaseRerunMap = new HashMap();
        List aBaseCTList = null;

        while ( iter.hasNext() ) {
            aRerun = (AudioArchiveBean) iter.next();
            aNumHalfHours
                = new Long( aRerun.getScheduledDuration() / (Util.MILLIS_IN_MINUTE * 30) );
            
            aCommonTimeList = (List) itsActiveRerunMap.get(aNumHalfHours);
            if ( aCommonTimeList == null ) {
                aCommonTimeList = new Vector();
                itsActiveRerunMap.put(aNumHalfHours, aCommonTimeList);
            }
            aCommonTimeList.add(aRerun);
        }

        // Sort all common times found in the map
        iter = itsActiveRerunMap.keySet().iterator();
        RerunComparator aRerunComparator = new RerunComparator();
        while ( iter.hasNext() ) {
            aNumHalfHours = (Long) iter.next();
            aCommonTimeList = (List) itsActiveRerunMap.get(aNumHalfHours);

            // XXX Might also consider doing a Collections.shuffle here...
            Collections.sort(aCommonTimeList, aRerunComparator);

            // Add a copy of each list to the base map
            aBaseCTList = new Vector();
            aBaseCTList.addAll(aCommonTimeList);
            itsBaseRerunMap.put(aNumHalfHours, aBaseCTList);
        }
    }

    private List getEmptySlots ( Connection theConn,
                                 int theChannelid ) throws SQLException {

        List anEmptySlotList = new Vector();
        PreparedStatement aPS = null;
        ResultSet aRS = null;
        String aSQL 
            = "SELECT * "
            + "FROM PublishedSchedule "
            + "WHERE starttime > ? " 
            + "OR ( endtime > ? AND starttime < ? ) "
            + "AND channelid = ? "
            + "ORDER BY starttime "
            ;

        // We can schedule reruns for times that are at least 5 hours in the future
        // XXX If we increase the maximum show length past 4 hours, we 
        // need to verify that this still works
        long aStartRerunSchedTime = Util.floorToHour(System.currentTimeMillis() + (Util.MILLIS_IN_HOUR * 5));
        debug.out("aStartRerunSchedTime=" + (new java.util.Date(aStartRerunSchedTime).toString()));
        
        // We can schedule reruns for times that are at most 42 hours in the future
        // (We could probably say 48, but let's be cautious here...)
        // XXX If we change the 48-hour advance schedule limit, we need to change this
        long anEndRerunSchedTime = Util.floorToHour(System.currentTimeMillis() + (Util.MILLIS_IN_HOUR * 42));

        long anEmptyStart = aStartRerunSchedTime;
        long aPubStart = -1;
        long aPubEnd = -1;
        EmptySlot anEmptySlot = null;

        try {
            aPS = theConn.prepareStatement(aSQL);
            int i=1;
            aPS.setLong(i++, aStartRerunSchedTime);
            aPS.setLong(i++, aStartRerunSchedTime);
            aPS.setLong(i++, anEndRerunSchedTime);
            aPS.setInt (i++, theChannelid);
            aRS = aPS.executeQuery();

            while ( aRS.next() ) {
                aPubStart = aRS.getLong("starttime");
                aPubEnd = aRS.getLong("endtime");

                if ( aPubStart - anEmptyStart >= (Util.MILLIS_IN_MINUTE*30) ) {
                    anEmptySlot = new EmptySlot(anEmptyStart, aPubStart);
                    anEmptySlotList.add(anEmptySlot);
                    debug.out("Found empty slot: " + anEmptySlot.toString());
                }
                anEmptyStart = aPubEnd;
                if ( anEmptyStart > anEndRerunSchedTime ) {
                    break;
                }
            }

            // Add a final empty slot if necessary
            if ( anEmptyStart < anEndRerunSchedTime ) {
                debug.out("anEmptyStart (in final) = " + (new java.util.Date(anEmptyStart)).toString());
                anEmptySlot = new EmptySlot(anEmptyStart, anEndRerunSchedTime);
                anEmptySlotList.add(anEmptySlot);
                debug.out("Found final empty slot: " + anEmptySlot.toString());
            }

        } finally {
            DBUtil.closeRS(aRS);
            DBUtil.closeStatement(aPS);
        }

        return anEmptySlotList;
    }

    private void fillSlotWithReruns ( Connection theConn, 
                                      EmptySlot theSlot ) throws SQLException {

        long aSlotDuration = theSlot.getDuration();
        long aFourHourMillis = (Util.MILLIS_IN_HOUR * 4);

        // If the slot is larger than 4 hours, split it into multiple 4-hour slots
        if ( aSlotDuration > aFourHourMillis ) {

            long aStartTime = theSlot.getStart();
            long anEndTime = theSlot.getEnd();
            EmptySlot aSlot = null;
            while ( aStartTime < anEndTime ) {
                if ( aStartTime + aFourHourMillis <= anEndTime ) {
                    aSlot = new EmptySlot(aStartTime, aStartTime + aFourHourMillis);
                    aStartTime += aFourHourMillis;
                } else {
                    aSlot = new EmptySlot(aStartTime, anEndTime);
                    aStartTime = anEndTime;
                }
                fillSlotWithReruns(theConn, aSlot);
            }
            return;
        }

        // Check for unfillable slots (should never happen)
        if ( aSlotDuration % (Util.MILLIS_IN_MINUTE * 30) != 0 ) {
            debug.out("Irregular empty slot, cannot fill: " + theSlot);
            return;
        }

        // Now, choose a fill strategy based on the slot size.
        itsStrategy.fillSlotWithReruns ( theConn, theSlot, this );
    }

    public boolean hasEnoughReruns( long theSize, int theCount, boolean theDoRegeneration ) {

        List aRerunList = null;
        List aNewList = null;
        Long aSizeLong = new Long(theSize/(Util.MILLIS_IN_MINUTE*30));

        debug.out("RerunScheduler.hasEnoughReruns(" + theSize + ", " 
                  + theCount + ", " + theDoRegeneration + ")");

        // See if we have any reruns of the specified size
        aRerunList = (List) itsActiveRerunMap.get(aSizeLong);

        if ( aRerunList == null || aRerunList.size() == 0 ) {

            if ( theDoRegeneration ) {
                // None left, pull list from base map
                aRerunList = (List) itsBaseRerunMap.get(aSizeLong);
                if ( aRerunList == null || aRerunList.size() == 0 ) {
                    // None ever existed
                    debug.out("hasEnoughReruns returning false");
                    return false;
                }

                // Put a copy of the base list back into the active list
                debug.out("hasEnoughReruns: REGENERATING FOR TIME=" + aSizeLong.longValue());
                aNewList = new Vector();
                aNewList.addAll(aRerunList);
                itsActiveRerunMap.put(aSizeLong, aNewList);
            } else {
                debug.out("hasEnoughReruns returning false (noregen)");
                return false;
            }
        }

        // Are there enough reruns in the list to use?
        debug.out("hasEnoughReruns returning " + String.valueOf(aRerunList.size() >= theCount));
        return (aRerunList.size() >= theCount);
    }

    public AudioArchiveBean getRerunOfLength( long theSize ) {

        List aRerunList = null;
        List aNewList = null;
        Long aSizeLong = new Long(theSize/(Util.MILLIS_IN_MINUTE*30));

        debug.out("RerunScheduler.getRerunOfLength(" + theSize + ")");

        // See if we have any reruns of the specified size
        aRerunList = (List) itsActiveRerunMap.get(aSizeLong);

        if ( aRerunList == null || aRerunList.size() == 0 ) {
            // None left, pull list from base map
            aRerunList = (List) itsBaseRerunMap.get(aSizeLong);
            if ( aRerunList == null || aRerunList.size() == 0 ) {
                // None ever existed
                debug.out("getRerunOfLength: Returning null");
                return null;
            }

            // Put a copy of the base list back into the active list
            aNewList = new Vector();
            aNewList.addAll(aRerunList);
            itsActiveRerunMap.put(aSizeLong, aNewList);
        }

        AudioArchiveBean aRerun = (AudioArchiveBean) aRerunList.remove(0);
        debug.out("getRerunOfLength: Returning " + aRerun);
        return aRerun;
    }

    public void assignReruns () {

        Connection aConn = null;
        PreparedStatement aPS = null;
        String aSQL 
            = "UPDATE PublishedSchedule "
            + "SET djid = ? "
            + "WHERE ( status = ? OR status = ? ) "
            + "AND starttime < ? "
            ;
        try {
            aConn = DBUtil.getConnection();
            aPS = aConn.prepareStatement(aSQL);
            int i=1;
            aPS.setInt (i++, RERUN_DJID);
            aPS.setInt (i++, SchedBlock.STATUS_DJ_SCHEDULED_RERUN);
            aPS.setInt (i++, SchedBlock.STATUS_NEEDSUB_WITH_RERUN);
            aPS.setLong(i++, (System.currentTimeMillis() + RERUN_SCHED_LIMIT_MILLIS));
            aPS.executeUpdate();

        } catch ( Exception e ) {
            debug.out("Unexpected exception in RerunScheduler.assignReruns: "
                      + Util.getStackTrace(e));
            LOG.capture("Unexpected exception in RerunScheduler.assignReruns", e, Logger.L_ERROR);

        } finally {
            DBUtil.closeStatement(aPS);
            DBUtil.closeConnection(aConn);
        }
    }
}
